﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Text;
using System.Windows.Forms;
// Jay (the Ham) Harland

namespace Nushi_Tsuri_KnB___Decompressor
{
    public partial class Form1 : Form
    {

        int enDataPosition;
        byte[] deData;
        byte[] sourceFile;

        public Form1()
        {
            InitializeComponent();
        }

        private void buttonFilePath_Click(object sender, EventArgs e)
        {
            OpenFileDialog openFile = new OpenFileDialog();

            if (openFile.ShowDialog() == DialogResult.OK)
            {
                textBoxSrcPath.Text = openFile.FileName.ToString();
                
                //Check the file can be opened
                try
                {
                    textBoxOutput.Clear();

                    FileStream readROM = new FileStream(textBoxSrcPath.Text, FileMode.Open);

                    readROM.Close();

                    buttonDecomp.Enabled = true;
                    buttonComp.Enabled = true;
                }
                catch (Exception x) 
                {
                    textBoxOutput.Text = Convert.ToString(x);
                }

            }
        }

        private void buttonDecomp_Click(object sender, EventArgs e)
        {
            bool offsetIsValid = false;

            //Convert the offset into an int
            try
            {
                textBoxOutput.Clear();

                enDataPosition = Int32.Parse(textBoxOffset.Text, System.Globalization.NumberStyles.HexNumber);
                offsetIsValid = true;

                FileStream readROM = new FileStream(textBoxSrcPath.Text, FileMode.Open);

                sourceFile = new byte[readROM.Length];

                readROM.Read(sourceFile, 0, sourceFile.Length);

                readROM.Close();
            }
            catch (Exception x)
            {
                textBoxOutput.Text = Convert.ToString(x);
            }

            if (offsetIsValid == true)
            {
                textBoxOutput.Text = String.Format("ROM size: {0:X8}, Data offset: {1:X8}", sourceFile.Length, enDataPosition);
                decoder();
            }
        }

        private void buttonComp_Click(object sender, EventArgs e)
        {
            encoder();
        }

        private void decoder()
        {
            int enFileOffset = enDataPosition;
            int deFileSize = sourceFile[enDataPosition] + (sourceFile[enDataPosition + 1] << 8);
            int enDataFlag = sourceFile[enDataPosition + 2];
            int dePosition = 0;
            enDataPosition += 3;//Increment offset to first data byte

            textBoxOutput.AppendText("\n");
            textBoxOutput.AppendText(String.Format("Decomp File size: {0:X8}", deFileSize));

            deData = new byte[deFileSize];
            
            while (dePosition != deFileSize)
            {
                //Check if raw data or compressed
                if (sourceFile[enDataPosition] == enDataFlag) //Compressed data
                {
                    enDataPosition++;

                    int distance = sourceFile[enDataPosition];
                    enDataPosition++;

                    byte numEncodedBytes = sourceFile[enDataPosition];
                    enDataPosition++;

                    distance = (ushort)(distance + ((numEncodedBytes & 0xf0)<< 4));

                    numEncodedBytes = (byte)((numEncodedBytes & 0x0f) + 4);

                    //Check if there's another num byte
                    if (numEncodedBytes == 0x13)
                    {
                        numEncodedBytes += sourceFile[enDataPosition];
                        enDataPosition++;
                    }


                    while (numEncodedBytes != 0)
                    {
                        if (distance >= deFileSize)
                            distance -= 0x1000;

                        if (distance < 0)
                            deData[dePosition] = 0x00;
                        else
                            deData[dePosition] = deData[distance];

                        dePosition++;
                        distance++;
                        numEncodedBytes--;

                    }
                }
                else //raw data
                {
                    deData[dePosition] = sourceFile[enDataPosition];

                    dePosition++;
                    enDataPosition++;
                }
            }

            textBoxOutput.AppendText("\n");
            textBoxOutput.AppendText(String.Format("End of compressed file: {0:X8}", enDataPosition));

            string outputFile = textBoxSrcPath.Text + "_decomp";
            FileStream writeOutput = new FileStream(outputFile, FileMode.Create);

            writeOutput.Write(deData, 0, deData.Length);
            writeOutput.Flush();
            writeOutput.Close();
        }

        private void encoder()
        {
            //Load the input file into a buffer array
            try
            {
                textBoxOutput.Clear();

                FileStream readROM = new FileStream(textBoxSrcPath.Text, FileMode.Open);

                sourceFile = new byte[readROM.Length];

                readROM.Read(sourceFile, 0, sourceFile.Length);

                readROM.Close();
            }
            catch (Exception x)
            {
                textBoxOutput.Text = Convert.ToString(x);
                return;
            }


            int srcPosition = 0;
            int compPosition = 0;
            int numBytesToEncode = 0;
            int readFromPosition;
            byte flag = 0;
            byte[] compData = new byte[sourceFile.Length * 2];


            //Save original file size in bytes as little endian
            compData[compPosition] = (byte)(sourceFile.Length);
            compPosition++;
            compData[compPosition] = (byte)(sourceFile.Length >> 8);
            compPosition++;


            //Calculate the compressed data flag
            bool[] valueTestArray = new bool[0x100];
            foreach (byte theByte in sourceFile)
            {
                //Mark as TRUE all used values
                valueTestArray[theByte] = true;
            }

            //The first unused value found will be the flag
            //If there is no flag available, warn and abort
            for (int value = 0; value < valueTestArray.Length; value++)
            {
                if (valueTestArray[value] == false)
                {
                    flag = (byte)value;
                    compData[2] = flag;
                    compPosition++;
                    break;
                }
                else if (value == 0xFF)
                {
                    textBoxOutput.AppendText("\n");
                    textBoxOutput.AppendText("*****");
                    textBoxOutput.AppendText("Data is not able to be compressed!");
                    textBoxOutput.AppendText("At least one value from the range 0x00-0xFF must NOT be present in the file");
                    textBoxOutput.AppendText("*****");
                    return;
                }
            }


            int comparePosition = 0;
            int numMatchingBytes = 0;


            //Encode all data
            //Search for longest match, not first found
            while (srcPosition < sourceFile.Length)
            {
                numBytesToEncode = 0;
                readFromPosition = 0;
                comparePosition = 0;
                numMatchingBytes = 0;

                while (comparePosition < srcPosition)
                {
                    while (sourceFile[srcPosition + numMatchingBytes] == sourceFile[comparePosition + numMatchingBytes])
                    {
                        numMatchingBytes++;

                        //Stop comparing if end of file is reached
                        if ((srcPosition + numMatchingBytes) >= sourceFile.Length)
                            break;

                        //Stop comparing if max number of compressable bytes is reached
                        if (numMatchingBytes == 0xFF)
                            break;

                    }
                    if (numMatchingBytes > numBytesToEncode)
                    {
                        numBytesToEncode = numMatchingBytes;
                        readFromPosition = comparePosition;
                    }

                    numMatchingBytes = 0;
                    comparePosition++;

                    //Only the first 0x1000 bytes are available for comparing
                    //I think
                    //or is it the first 0x7FF
                    //or something different?!
                    if (comparePosition == 0x1000)
                        break;
                }


                //Compressing 0x12 instead of 0x13 will save a byte being wasted on extended count
                if (numBytesToEncode == 0x13)
                    numBytesToEncode = 0x12;
                

                if (numBytesToEncode > 3) //If there are enough bytes matching to compress
                {
                    //Write flag
                    compData[compPosition] = flag;
                    compPosition++;

                    //Write lowbyte of readFromPosition
                    compData[compPosition] = (byte)(readFromPosition & 0xFF);
                    compPosition++;

                    //Write high nibble of readFromPosition
                    readFromPosition = readFromPosition & 0xF00;
                    readFromPosition = readFromPosition >> 4;
                    compData[compPosition] = (byte)readFromPosition;


                    //Extend number to copy, across two bytes if needed
                    if (numBytesToEncode > 0x12)
                    {
                        compData[compPosition] = (byte)(compData[compPosition] + 0x0f);
                        compPosition++;
                        compData[compPosition] = (byte)(numBytesToEncode - 0x13);
                    }
                    else
                    {
                        compData[compPosition] = (byte)(compData[compPosition] + (numBytesToEncode - 4));
                    }
                    compPosition++;
                    srcPosition += numBytesToEncode;

                }
                else //It's just a single byte copy
                {
                    compData[compPosition] = sourceFile[srcPosition];
                    compPosition++;
                    srcPosition++;
                }
            }


            string outputFile = textBoxSrcPath.Text + "_comp";
            FileStream writeOutput = new FileStream(outputFile, FileMode.Create);

            writeOutput.Write(compData, 0, compPosition);
            writeOutput.Flush();
            writeOutput.Close();
        }


    }
}
